from functools import partial

import gdsfactory as gf
from gdsfactory import cross_section as xs
from gdsfactory.generic_tech.layer_map import LAYER

###################
# Constants
###################
pad_size = (80, 80)
pad_spacing = 100

###################
# Cells
###################
straight_sc = partial(gf.c.straight, cross_section=xs.strip)
straight_heater_metal = partial(
    gf.c.straight_heater_metal,
    cross_section=xs.strip,
)
terminator = partial(gf.c.taper, width2=1)
grating_coupler_te = partial(gf.c.grating_coupler_te, cross_section=xs.strip)
pad = partial(gf.c.pad, size=pad_size, layer=LAYER.MTOP)

mmi1x2 = partial(
    gf.c.mmi1x2,
    width_taper=1,
    length_taper=10,
    length_mmi=10,
    width_mmi=2.5,
    cross_section=xs.strip,
)

#################################
# Cells that contain other cells
#################################

add_fiber_array_optical_south_electrical_north = partial(
    gf.c.add_fiber_array_optical_south_electrical_north,
    component=straight_heater_metal,
    pad=pad,
    grating_coupler=grating_coupler_te,
    cross_section_metal=xs.metal_routing,
    pad_spacing=pad_spacing,
)
add_termination = partial(gf.c.add_termination, terminator=terminator)


if __name__ == "__main__":
    c = add_fiber_array_optical_south_electrical_north()
    c.show()



from pydantic.dataclasses import dataclass

from gdsfactory.generic_tech.layer_map import LAYER
from gdsfactory.technology import LayerLevel, LayerStack, LogicalLayer
from gdsfactory.technology.processes import (
    Anneal,
    Etch,
    Grow,
    ImplantPhysical,
    Planarize,
)

nm = 1e-3


class LayerStackParameters:
    """values used by get_layer_stack and get_process."""

    thickness_wg: float = 220 * nm
    thickness_slab_deep_etch: float = 90 * nm
    thickness_slab_shallow_etch: float = 150 * nm
    sidewall_angle_wg: float = 10
    thickness_clad: float = 3.0
    thickness_nitride: float = 350 * nm
    thickness_ge: float = 500 * nm
    gap_silicon_to_nitride: float = 100 * nm
    zmin_heater: float = 1.1
    zmin_metal1: float = 1.1
    thickness_metal1: float = 700 * nm
    zmin_metal2: float = 2.3
    thickness_metal2: float = 700 * nm
    zmin_metal3: float = 3.2
    thickness_metal3: float = 2000 * nm
    substrate_thickness: float = 10.0
    box_thickness: float = 3.0
    undercut_thickness: float = 5.0


def get_layer_stack(
    thickness_wg=LayerStackParameters.thickness_wg,
    thickness_slab_deep_etch=LayerStackParameters.thickness_slab_deep_etch,
    thickness_slab_shallow_etch=LayerStackParameters.thickness_slab_shallow_etch,
    sidewall_angle_wg=LayerStackParameters.sidewall_angle_wg,
    thickness_clad=LayerStackParameters.thickness_clad,
    thickness_nitride=LayerStackParameters.thickness_nitride,
    thickness_ge=LayerStackParameters.thickness_ge,
    gap_silicon_to_nitride=LayerStackParameters.gap_silicon_to_nitride,
    zmin_heater=LayerStackParameters.zmin_heater,
    zmin_metal1=LayerStackParameters.zmin_metal1,
    thickness_metal1=LayerStackParameters.thickness_metal1,
    zmin_metal2=LayerStackParameters.zmin_metal2,
    thickness_metal2=LayerStackParameters.thickness_metal2,
    zmin_metal3=LayerStackParameters.zmin_metal3,
    thickness_metal3=LayerStackParameters.thickness_metal3,
    substrate_thickness=LayerStackParameters.substrate_thickness,
    box_thickness=LayerStackParameters.box_thickness,
    undercut_thickness=LayerStackParameters.undercut_thickness,
    layer_wafer=LogicalLayer(layer=LAYER.WAFER),
    layer_core=LogicalLayer(layer=LAYER.WG),
    layer_shallow_etch=LogicalLayer(layer=LAYER.SHALLOW_ETCH),
    layer_deep_etch=LogicalLayer(layer=LAYER.DEEP_ETCH),
    layer_nitride=LogicalLayer(layer=LAYER.WGN),
    layer_slab_deep_etch=LogicalLayer(layer=LAYER.SLAB90),
    layer_slab_shallow_etch=LogicalLayer(layer=LAYER.SLAB150),
    layer_ge=LogicalLayer(layer=LAYER.GE),
    layer_undercut=LogicalLayer(layer=LAYER.UNDERCUT),
    layer_heater=LogicalLayer(layer=LAYER.HEATER),
    layer_metal1=LogicalLayer(layer=LAYER.M1),
    layer_metal2=LogicalLayer(layer=LAYER.M2),
    layer_metal3=LogicalLayer(layer=LAYER.M3),
    layer_viac=LogicalLayer(layer=LAYER.VIAC),
    layer_via1=LogicalLayer(layer=LAYER.VIA1),
    layer_via2=LogicalLayer(layer=LAYER.VIA2),
) -> LayerStack:
    """Returns generic LayerStack.

    based on paper https://www.degruyter.com/document/doi/10.1515/nanoph-2013-0034/html

    Args:
        thickness_wg: waveguide thickness in um.
        thickness_slab_deep_etch: for deep etched slab.
        thickness_slab_shallow_etch: thickness for the etch in um.
        sidewall_angle_wg: waveguide side angle.
        thickness_clad: cladding thickness in um.
        thickness_nitride: nitride thickness in um.
        thickness_ge: germanium thickness.
        gap_silicon_to_nitride: distance from silicon to nitride in um.
        zmin_heater: TiN heater.
        zmin_metal1: metal1.
        thickness_metal1: metal1 thickness.
        zmin_metal2: metal2.
        thickness_metal2: metal2 thickness.
        zmin_metal3: metal3.
        thickness_metal3: metal3 thickness.
        substrate_thickness: substrate thickness in um.
        box_thickness: bottom oxide thickness in um.
        undercut_thickness: thickness of the silicon undercut.
        layer_wafer: wafer layer.
        layer_core: waveguide layer.
        layer_shallow_etch: shallow etch layer.
        layer_deep_etch: deep etch layer.
        layer_nitride: nitride layer.
        layer_slab_deep_etch: deep etch slab layer.
        layer_slab_shallow_etch: shallow etch slab layer.
        layer_ge: germanium layer.
        layer_undercut: undercut layer.
        layer_heater: heater layer.
        layer_metal1: metal1 layer.
        layer_metal2: metal2 layer.
        layer_metal3: metal3 layer.
        layer_viac: viac layer.
        layer_via1: via1 layer.
        layer_via2: via2 layer.

    """
    thickness_deep_etch = thickness_wg - thickness_slab_deep_etch
    thickness_shallow_etch = thickness_wg - thickness_slab_shallow_etch
    layers = dict(
        substrate=LayerLevel(
            layer=layer_wafer,
            thickness=substrate_thickness,
            zmin=-substrate_thickness - box_thickness,
            material="si",
            mesh_order=101,
        ),
        box=LayerLevel(
            layer=layer_wafer,
            thickness=box_thickness,
            zmin=-box_thickness,
            material="sio2",
            mesh_order=9,
        ),
        core=LayerLevel(
            layer=layer_core - layer_deep_etch - layer_shallow_etch,
            thickness=thickness_wg,
            zmin=0.0,
            material="si",
            mesh_order=2,
            sidewall_angle=sidewall_angle_wg,
            width_to_z=0.5,
            derived_layer=LogicalLayer(layer=LAYER.WG),
        ),
        shallow_etch=LayerLevel(
            layer=layer_shallow_etch & layer_core,
            thickness=thickness_shallow_etch,
            zmin=0.0,
            material="si",
            mesh_order=1,
            derived_layer=LogicalLayer(layer=LAYER.SLAB150),
        ),
        deep_etch=LayerLevel(
            layer=layer_deep_etch,
            thickness=thickness_deep_etch,
            zmin=0.0,
            material="si",
            mesh_order=1,
            derived_layer=layer_slab_deep_etch,
        ),
        clad=LayerLevel(
            layer=layer_wafer,
            zmin=0.0,
            material="sio2",
            thickness=thickness_clad,
            mesh_order=10,
        ),
        slab150=LayerLevel(
            layer=layer_slab_shallow_etch,
            thickness=150e-3,
            zmin=0,
            material="si",
            mesh_order=3,
        ),
        slab90=LayerLevel(
            layer=layer_slab_deep_etch,
            thickness=thickness_slab_deep_etch,
            zmin=0.0,
            material="si",
            mesh_order=2,
        ),
        nitride=LayerLevel(
            layer=layer_nitride,
            thickness=thickness_nitride,
            zmin=thickness_wg + gap_silicon_to_nitride,
            material="sin",
            mesh_order=2,
        ),
        ge=LayerLevel(
            layer=layer_ge,
            thickness=thickness_ge,
            zmin=thickness_wg,
            material="ge",
            mesh_order=1,
        ),
        undercut=LayerLevel(
            layer=layer_undercut,
            thickness=-undercut_thickness,
            zmin=-box_thickness,
            material="air",
            z_to_bias=(
                [0, 0.3, 0.6, 0.8, 0.9, 1],
                [-0, -0.5, -1, -1.5, -2, -2.5],
            ),
            mesh_order=1,
        ),
        via_contact=LayerLevel(
            layer=layer_viac,
            thickness=zmin_metal1 - thickness_slab_deep_etch,
            zmin=thickness_slab_deep_etch,
            material="Aluminum",
            mesh_order=1,
            sidewall_angle=-10,
            width_to_z=0,
        ),
        metal1=LayerLevel(
            layer=layer_metal1,
            thickness=thickness_metal1,
            zmin=zmin_metal1,
            material="Aluminum",
            mesh_order=2,
        ),
        heater=LayerLevel(
            layer=layer_heater,
            thickness=750e-3,
            zmin=zmin_heater,
            material="TiN",
            mesh_order=2,
        ),
        via1=LayerLevel(
            layer=layer_via1,
            thickness=zmin_metal2 - (zmin_metal1 + thickness_metal1),
            zmin=zmin_metal1 + thickness_metal1,
            material="Aluminum",
            mesh_order=1,
        ),
        metal2=LayerLevel(
            layer=layer_metal2,
            thickness=thickness_metal2,
            zmin=zmin_metal2,
            material="Aluminum",
            mesh_order=2,
        ),
        via2=LayerLevel(
            layer=layer_via2,
            thickness=zmin_metal3 - (zmin_metal2 + thickness_metal2),
            zmin=zmin_metal2 + thickness_metal2,
            material="Aluminum",
            mesh_order=1,
        ),
        metal3=LayerLevel(
            layer=layer_metal3,
            thickness=thickness_metal3,
            zmin=zmin_metal3,
            material="Aluminum",
            mesh_order=2,
        ),
    )

    return LayerStack(layers=layers)


LAYER_STACK = get_layer_stack()


WAFER_STACK = LayerStack(
    layers={
        k: get_layer_stack().layers[k]
        for k in (
            "substrate",
            "box",
            "core",
        )
    }
)


def get_process() -> tuple[dataclass, ...]:
    """Returns generic process to generate LayerStack.

    Represents processing steps that will result in the GenericLayerStack, starting from the waferstack LayerStack.

    based on paper https://www.degruyter.com/document/doi/10.1515/nanoph-2013-0034/html
    """
    return (
        Etch(
            name="strip_etch",
            layer=LAYER.WG,
            layers_or=[LAYER.SLAB90],
            depth=LayerStackParameters.thickness_wg
            + 0.01,  # slight overetch for numerics
            material="silicon",
            resist_thickness=1.0,
            positive_tone=False,
        ),
        Etch(
            name="slab_etch",
            layer=LAYER.SLAB90,
            layers_diff=[LAYER.WG],
            depth=LayerStackParameters.thickness_wg
            - LayerStackParameters.thickness_slab_deep_etch,
            material="silicon",
            resist_thickness=1.0,
        ),
        # See gplugins.process.implant tables for ballpark numbers
        # Adjust to your process
        ImplantPhysical(
            name="deep_n_implant",
            layer=LAYER.N,
            energy=100,
            ion="P",
            dose=1e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="shallow_n_implant",
            layer=LAYER.N,
            energy=50,
            ion="P",
            dose=1e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="deep_p_implant",
            layer=LAYER.P,
            energy=50,
            ion="B",
            dose=1e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="shallow_p_implant",
            layer=LAYER.P,
            energy=15,
            ion="B",
            dose=1e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="pp_implant",
            layer=LAYER.PP,
            energy=15,
            ion="B",
            dose=5e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="np_implant",
            layer=LAYER.NP,
            energy=50,
            ion="P",
            dose=5e12,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="ppp_implant",
            layer=LAYER.PPP,
            energy=15,
            ion="B",
            dose=1e15,
            resist_thickness=1.0,
        ),
        ImplantPhysical(
            name="npp_implant",
            layer=LAYER.NPP,
            energy=100,
            ion="As",
            dose=1e15,
            resist_thickness=1.0,
        ),
        # "Temperatures of ~1000C for not more than a few seconds"
        # Adjust to your process
        # https://en.wikipedia.org/wiki/Rapid_thermal_processing
        Anneal(
            name="dopant_activation",
            time=5,
            temperature=1000,
        ),
        Grow(
            name="viac_metallization",
            layer=None,
            thickness=LayerStackParameters.zmin_metal1
            - LayerStackParameters.thickness_slab_deep_etch,
            material="Aluminum",
            type="anisotropic",
        ),
        Etch(
            name="viac_etch",
            layer=LAYER.VIAC,
            depth=LayerStackParameters.zmin_metal1
            - LayerStackParameters.thickness_slab_deep_etch
            + 0.1,
            material="Aluminum",
            type="anisotropic",
            resist_thickness=1.0,
            positive_tone=False,
        ),
        Grow(
            name="deposit_cladding",
            layer=None,
            thickness=LayerStackParameters.thickness_clad
            + LayerStackParameters.thickness_slab_deep_etch,
            material="Oxide",
            type="anisotropic",
        ),
        Planarize(
            name="planarization",
            height=LayerStackParameters.thickness_clad
            - LayerStackParameters.thickness_slab_deep_etch,
        ),
    )


if __name__ == "__main__":
    # ls = get_layer_stack(substrate_thickness=50.0)
    # ls = get_layer_stack()
    # script = ls.get_klayout_3d_script()
    # print(script)
    # print(ls.get_layer_to_material())
    # print(ls.get_layer_to_thickness())

    for layername, layer in WAFER_STACK.layers.items():
        print(layername, layer.zmin, layer.thickness)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.typings import LayerSpec


def bbox_to_points(
    bbox: gf.kdb.DBox,
    top: float = 0,
    bottom: float = 0,
    left: float = 0,
    right: float = 0,
) -> list[tuple[float, float]]:
    """Returns bounding box rectangle with offsets.

    Args:
        bbox: DBbox.
        top: north offset.
        bottom: south offset.
        left: west offset.
        right: east offset.

    """
    xmin, ymin, xmax, ymax = bbox.left, bbox.bottom, bbox.right, bbox.top
    xmin = float(xmin)
    xmax = float(xmax)
    ymin = float(ymin)
    ymax = float(ymax)
    return [
        (xmin - left, ymin - bottom),
        (xmax + right, ymin - bottom),
        (xmax + right, ymax + top),
        (xmin - left, ymax + top),
    ]


@gf.cell
def bbox(
    component: gf.Component | gf.Instance,
    layer: LayerSpec,
    top: float = 0,
    bottom: float = 0,
    left: float = 0,
    right: float = 0,
) -> gf.Component:
    """Returns bounding box rectangle from coordinates.

    Args:
        component: component or instance to get bbox from.
        layer: for bbox.
        top: north offset.
        bottom: south offset.
        left: west offset.
        right: east offset.
    """
    c = gf.Component()
    bbox = component.dbbox()
    xmin, ymin, xmax, ymax = bbox.left, bbox.bottom, bbox.right, bbox.top
    points = [
        [xmin - left, ymin - bottom],
        [xmax + right, ymin - bottom],
        [xmax + right, ymax + top],
        [xmin - left, ymax + top],
    ]
    c.add_polygon(points, layer=layer)
    return c


if __name__ == "__main__":
    # c = gf.Component()
    # l= gf.components.L()
    r = gf.c.text()
    b = bbox(r, layer=(2, 0))
    c = gf.Component()
    c << r
    c << b
    c.show()



from __future__ import annotations

from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.components.via import via
from gdsfactory.components.via_stack import via_stack
from gdsfactory.cross_section import Section
from gdsfactory.typings import ComponentSpec, CrossSection, LayerSpec

_pn_cross_section = partial(
    gf.cross_section.pn,
    width_doping=2.425,
    width_slab=2 * 2.425,
    layer_via="VIAC",
    width_via=0.5,
    layer_metal="M1",
    width_metal=0.5,
)

_heater_vias = partial(
    via_stack,
    size=(0.5, 0.5),
    layers=("M1", "M2"),
    vias=(
        partial(
            via,
            layer="VIAC",
            size=(0.1, 0.1),
            spacing=(0.2, 0.2),
            enclosure=0.1,
        ),
        partial(
            via,
            layer="VIA1",
            size=(0.1, 0.1),
            spacing=(0.2, 0.2),
            enclosure=0.1,
        ),
    ),
)

_cross_section = partial(
    gf.cross_section.strip,
    sections=(Section(width=2 * 2.425, layer="SLAB90", name="slab"),),
)


@gf.cell
def ring_single_pn(
    gap: float = 0.3,
    radius: float = 5.0,
    doping_angle: float = 250,
    cross_section: CrossSection = _cross_section,
    pn_cross_section: CrossSection = _pn_cross_section,
    doped_heater: bool = True,
    doped_heater_angle_buffer: float = 10,
    doped_heater_layer: LayerSpec = "NPP",
    doped_heater_width: float = 0.5,
    doped_heater_waveguide_offset: float = 2.175,
    heater_vias: ComponentSpec = _heater_vias,
) -> gf.Component:
    """Returns single pn ring with optional doped heater.

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        doping_angle: angle in degrees representing portion of ring that is doped.
        length_x: ring coupler length.
        length_y: vertical straight length.
        cross_section: cross_section spec for non-PN doped rib waveguide sections.
        pn_cross_section: cross section of pn junction.
        doped_heater: boolean for if we include doped heater or not.
        doped_heater_angle_buffer: angle in degrees buffering heater from pn junction.
        doped_heater_layer: doping layer for heater.
        doped_heater_width: width of doped heater.
        doped_heater_waveguide_offset: distance from the center of the ring waveguide to the center of the doped heater.
        heater_vias: components specifications for heater vias.
    """
    gap = gf.snap.snap_to_grid(gap, grid_factor=2)
    c = gf.Component()

    undoping_angle = 360 - doping_angle

    pn_cross_section = gf.get_cross_section(pn_cross_section)
    bus_waveguide_path = gf.Path()
    bus_waveguide_path.append(
        gf.path.straight(length=2 * radius * np.sin(np.pi / 360 * undoping_angle))
    )
    bus_waveguide = c << bus_waveguide_path.extrude(cross_section=cross_section)
    bus_waveguide.dx = 0
    bus_waveguide.dy = (
        -radius
        - gap
        - bus_waveguide.ports["o1"].dwidth / 2
        - pn_cross_section.width / 2
    )

    r = gf.ComponentAllAngle()
    doped_path = gf.Path()
    doped_path.append(gf.path.arc(radius=radius, angle=-doping_angle))
    undoped_path = gf.Path()
    undoped_path.append(gf.path.arc(radius=radius, angle=undoping_angle))

    doped_ring_ref = r << doped_path.extrude(
        cross_section=pn_cross_section, all_angle=True
    )
    undoped_ring_ref = r << undoped_path.extrude(
        cross_section=cross_section, all_angle=True
    )
    undoped_ring_ref.drotate(-undoping_angle / 2)
    undoped_ring_ref.dcenter = (0, 0)
    doped_ring_ref.connect("o1", undoped_ring_ref.ports["o1"])
    ring = c.create_vinst(r)
    ring.dcenter = (0, 0)

    if doped_heater:
        heater_radius = radius - doped_heater_waveguide_offset
        heater_path = gf.Path()
        heater_path.append(
            gf.path.arc(
                radius=heater_radius, angle=undoping_angle - doped_heater_angle_buffer
            )
        )

        bottom_heater_ref = c << heater_path.extrude(
            width=0.5, layer=doped_heater_layer
        )
        bottom_heater_ref.drotate(-(undoping_angle - doped_heater_angle_buffer) / 2)
        bottom_heater_ref.dx = bus_waveguide.dx
        bottom_heater_ref.dy = (
            bus_waveguide.dy
            + doped_heater_waveguide_offset
            + doped_heater_width / 2
            + gap
        )

        heater_vias = gf.get_component(heater_vias)

        bottom_l_heater_via = c << heater_vias
        bottom_r_heater_via = c << heater_vias
        bottom_l_heater_via.connect(
            "e3",
            bottom_heater_ref.ports["o1"],
            allow_layer_mismatch=True,
            allow_type_mismatch=True,
        )
        bottom_r_heater_via.connect(
            "e3",
            bottom_heater_ref.ports["o2"],
            allow_layer_mismatch=True,
            allow_type_mismatch=True,
        )

    c.add_port("o1", port=bus_waveguide.ports["o1"])
    c.add_port("o2", port=bus_waveguide.ports["o2"])
    c.flatten()
    return c


if __name__ == "__main__":
    # c = ring_single(layer=(2, 0), cross_section_factory=gf.cross_section.pin, width=1)
    # c = ring_single(width=2, gap=1, layer=(2, 0), radius=7, length_y=1)
    # print(c.ports)

    # c = gf.routing.add_fiber_array(ring_single)
    c = ring_single_pn()
    c.show()

    # cc = gf.add_pins(c)
    # print(c.settings)
    # print(c.settings)
    # cc.show( )



from __future__ import annotations

import gdsfactory as gf
from gdsfactory import cell
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.routing.route_single import route_single
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def loss_deembedding_ch13_24(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = "grating_coupler_te",
    cross_section: CrossSectionSpec = "strip",
    port_name: str = "o1",
    rotation: float = -90,
    yspacing: float | None = None,
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->3, 2->4

    Args:
        pitch: um.
        grating_coupler: spec.
        cross_section: spec.
        port_name: for the grating_coupler port.
        rotation: degrees.
        yspacing: um.
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)
    c = gf.Component()
    dx = pitch
    gc_ports = []

    for i in range(4):
        g = c << gc
        g.drotate(rotation)
        g.dmovex(i * dx)
        gc_ports.append(g.ports[port_name])

    route_single(
        c,
        gc_ports[0],
        gc_ports[2],
        start_straight_length=40.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )

    x = gf.get_cross_section(cross_section, **kwargs)
    radius = x.radius

    p1 = gc_ports[1]
    p3 = gc_ports[3]
    yspacing = yspacing or gc.dysize + 2 * radius
    bend90 = gf.components.bend_euler(cross_section=cross_section, **kwargs)
    points = gf.kf.routing.optical.route_loopback(
        p1,
        p3,
        bend90_radius=round(radius / c.kcl.dbu),
        d_loop=round(yspacing / c.kcl.dbu),
    )
    route_single(
        c,
        port1=p1,
        port2=p3,
        waypoints=points,
        bend=bend90,
        straight=gf.components.straight,
        cross_section=cross_section,
        **kwargs,
    )

    return c


@cell
def loss_deembedding_ch12_34(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = "grating_coupler_te",
    port_name: str = "o1",
    cross_section: CrossSectionSpec = "strip",
    rotation: float = -90,
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->2, 3->4

    Args:
        pitch: um.
        grating_coupler: spec.
        port_name: for the grating_coupler port.
        cross_section: spec.
        rotation: degrees.
        kwargs: cross_section settings.

    Keyword Args:
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch

    gc_ports = []

    for i in range(4):
        g = c << gc
        g.drotate(rotation)
        g.dmove((i * dx, 0))
        gc_ports.append(g.ports[port_name])

    route_single(
        c,
        gc_ports[0],
        gc_ports[1],
        start_straight_length=40.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )
    route_single(
        c,
        gc_ports[2],
        gc_ports[3],
        start_straight_length=40.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )
    return c


@cell
def loss_deembedding_ch14_23(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = "grating_coupler_te",
    cross_section: CrossSectionSpec = "strip",
    port_name: str = "o1",
    rotation: float = -90,
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->4, 2->3

    Args:
        pitch: um.
        grating_coupler: spec.
        cross_section: spec.
        port_name: for the grating_coupler port.
        rotation: degrees.
        kwargs: cross_section settings.

    Keyword Args:
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch
    gc_ports = []

    for i in range(4):
        g = c << gc
        g.drotate(rotation)
        g.dmove((i * dx, 0))
        gc_ports.append(g.ports[port_name])

    route_single(
        c,
        gc_ports[0],
        gc_ports[3],
        start_straight_length=40.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )
    route_single(
        c,
        gc_ports[1],
        gc_ports[2],
        start_straight_length=30.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )
    return c


@cell
def grating_coupler_loss_fiber_array(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = "grating_coupler_te",
    port_name: str = "o1",
    cross_section: CrossSectionSpec = "strip",
    rotation: float = -90,
    **kwargs,
) -> Component:
    """Returns Grating coupler fiber array loopback.

    Args:
        pitch: spacing.
        grating_coupler: spec for grating coupler.
        port_name: for the grating_coupler port.
        cross_section: spec.
        rotation: degrees.
        kwargs: cross_section settings.

    Keyword Args:
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch
    gc_ports = []

    for i in range(2):
        g = c << gc
        g.drotate(rotation)
        g.dmove((i * dx, 0))
        gc_ports.append(g.ports[port_name])

    route_single(
        c,
        gc_ports[0],
        gc_ports[1],
        start_straight_length=40.0,
        taper=None,
        cross_section=cross_section,
        **kwargs,
    )
    return c


@cell
def grating_coupler_loss_fiber_array4(
    pitch: float = 127.0, grating_coupler: ComponentSpec = grating_coupler_te, **kwargs
) -> Component:
    """Returns a grating coupler test structure for fiber array.

    Measures all combinations for a 4 fiber fiber_array

    Connects channel 1->3, 2->4
    Connects channel 1->4, 2->3
    Connects channel 1->2, 3->4

    Args:
        pitch: grating_coupler_pitch.
        grating_coupler: function.
        kwargs: cross_section settings.
    """
    c = gf.Component()
    c1 = loss_deembedding_ch13_24(grating_coupler=grating_coupler, **kwargs)
    c2 = loss_deembedding_ch14_23(grating_coupler=grating_coupler, **kwargs)
    c3 = loss_deembedding_ch12_34(grating_coupler=grating_coupler, **kwargs)
    c.add_ref(c1)
    c2 = c.add_ref(c2)
    c3 = c.add_ref(c3)
    c2.dmovex(pitch * 4)
    c3.dmovex(pitch * 8)
    return c


if __name__ == "__main__":
    # c = loss_deembedding_ch14_23()
    # c = loss_deembedding_ch12_34()
    # c = loss_deembedding_ch13_24()
    # c = grating_coupler_loss_fiber_array4()
    # c = grating_coupler_loss_fiber_array4(layer=(2, 0), radius=30)
    # c = grating_coupler_loss_fiber_array4(cross_section="rib")
    # c = grating_coupler_loss_fiber_array(layer=(2, 0), radius=30)
    c = grating_coupler_loss_fiber_array()
    c.show()



"""Straight Doped PIN waveguide."""

from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper_strip_to_ridge
from gdsfactory.components.via_stack import via_stack_slab_m3
from gdsfactory.cross_section import pin
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def straight_pin(
    length: float = 500.0,
    cross_section: CrossSectionSpec = pin,
    via_stack: ComponentSpec = via_stack_slab_m3,
    via_stack_width: float = 10.0,
    via_stack_spacing: float = 2,
    taper: ComponentSpec | None = taper_strip_to_ridge,
) -> Component:
    """Returns rib waveguide with doping and via_stacks used for PN and PIN modulators.

    For PIN:
    https://doi.org/10.1364/OE.26.029983

    500um length for PI phase shift
    https://ieeexplore.ieee.org/document/8268112

    to go beyond 2PI, you will need at least 1mm
    https://ieeexplore.ieee.org/document/8853396/

    For PN:
    Typical lengths in practice are 2-5mm depending on doping,engineering and application:
    https://opg.optica.org/oe/fulltext.cfm?uri=oe-21-25-30350&id=275107
    https://opg.optica.org/oe/fulltext.cfm?uri=oe-20-11-12014&id=233271

    Args:
        length: of the waveguide.
        cross_section: for the waveguide.
        via_stack: for the via_stacks.
        via_stack_width: width of the via_stack.
        via_stack_spacing: spacing between via_stacks.
        taper: optional taper.
    """
    c = Component()
    if taper:
        _taper = gf.get_component(taper)
        length -= 2 * _taper.dxsize

    wg = c << gf.components.straight(
        cross_section=cross_section,
        length=length,
    )

    if taper:
        t1 = c << _taper
        t2 = c << _taper
        t1.connect("o2", wg.ports["o1"])
        t2.connect("o2", wg.ports["o2"])
        c.add_port("o1", port=t1.ports["o1"])
        c.add_port("o2", port=t2.ports["o1"])

    else:
        c.add_ports(wg.ports)

    via_stack_length = length
    _via_stack = gf.get_component(via_stack, size=(via_stack_length, via_stack_width))
    via_stack_top = c << _via_stack
    via_stack_bot = c << _via_stack
    via_stack_bot.dxmin = wg.dxmin
    via_stack_top.dxmin = wg.dxmin

    # via_stack_top.dymin = +int(via_stack_spacing / 2 / c.kcl.dbu)
    # via_stack_bot.dymax = -int(via_stack_spacing / 2 / c.kcl.dbu)

    via_stack_top.dymin = +via_stack_spacing / 2
    via_stack_bot.dymax = -via_stack_spacing / 2

    c.add_ports(via_stack_bot.ports, prefix="bot_")
    c.add_ports(via_stack_top.ports, prefix="top_")
    return c


straight_pn = partial(straight_pin, cross_section="pn", length=2000)

if __name__ == "__main__":
    c = straight_pn(length=40)
    c.show()



"""Add loopback reference for a grating coupler array."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.routing.route_single import route_single
from gdsfactory.typings import Component, ComponentSpec


@gf.cell
def add_loopback(
    component: ComponentSpec,
    port1_name: str,
    port2_name: str,
    grating: ComponentSpec,
    grating_separation: float = 127.0,
    grating_rotation: int = -90,
    grating_port_name: str = "o1",
    bend: ComponentSpec = bend_euler,
    south_waveguide_spacing: float | None = None,
    inside: bool = True,
    **kwargs,
) -> Component:
    """Return loopback (grating coupler align reference) references.

    Input grating generated on the left of port1
    Output grating generated on the right of port2

    Args:
        component: input component.
        port1_name: start port.
        port2_name: end port.
        grating: fiber coupler.
        grating_separation: grating pitch.
        grating_rotation: in degrees.
        grating_port_name: fiber port name for grating coupler.
        bend: bend spec.
        south_waveguide_spacing: spacing from loopback to grating_coupler.dymin
        inside: add loopback inside.
        kwargs: cross_section settings.

    .. code::

    inside = True
         ______                     ______
        |<-separation  |     |      |     |
        |      |       |     |      |     |
       GC      |    port1  port2    |    GC___
               |                    |       |
               |                    |       | south_waveguide_spacing
               |____________________|      _|_

    inside = False
                ______                    _______
               |     |                   |      |
               |     |       |     |     |      |
               |     GC   port1  port2   GC     |      ___
               |                                |       |
               |                                |       | south_waveguide_spacing
               |________________________________|      _|_
    """
    port1 = component.ports[port1_name]
    port2 = component.ports[port2_name]

    c = gf.Component()
    gc = gf.get_component(grating)

    y0 = port1.dy
    x0 = port1.dx - grating_separation
    x1 = port2.dx + grating_separation

    gca1 = c << gc
    gca1.drotate(grating_rotation)
    gca1.dmove((x0, y0))

    gca2 = c << gc
    gca2.drotate(grating_rotation)
    gca2.dmove((x1, y0))

    p0 = gca1.ports[grating_port_name].dcenter
    p1 = gca2.ports[grating_port_name].dcenter
    bend90 = bend(**kwargs)

    a = abs(bend90.info["dy"]) if hasattr(bend90, "dx") else bend90.dxsize + 0.5
    b = max(2 * a, grating_separation / 2)
    b = b if inside else -b

    south_waveguide_spacing = (
        south_waveguide_spacing
        if south_waveguide_spacing is not None
        else -gc.dxsize - 5.0
    )

    waypoints = [
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, south_waveguide_spacing),
        p1 + (-b, south_waveguide_spacing),
        p1 + (-b, a),
        p1 + (0, a),
    ]
    route_single(
        c, port1=port1, port2=port2, waypoints=waypoints, bend=bend90, **kwargs
    )
    return c


if __name__ == "__main__":
    wg = gf.components.straight()
    c = add_loopback(
        wg,
        "o1",
        "o2",
        grating=gf.components.grating_coupler_te,
        inside=False,
    )
    c.show()



from __future__ import annotations

import itertools as it
from typing import Any

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.grid import grid, grid_with_text
from gdsfactory.pack import pack
from gdsfactory.typings import CellSpec, ComponentSpec

_doe = "mmi1x2"
_settings = dict(length_mmi=(2.5, 100), width_mmi=(4, 10))


def generate_doe(
    doe: ComponentSpec,
    settings: dict[str, list[Any]],
    do_permutations: bool = False,
    function: CellSpec | None = None,
) -> tuple[tuple[Component, ...], tuple[dict, ...]]:
    """Generates a component DOE (Design of Experiment).

    which can then be packed, or used elsewhere.

    Args:
        doe: function to return Components.
        settings: component settings.
        do_permutations: for each setting.
        function: for the component (add padding, grating couplers ...)
    """
    if do_permutations:
        settings_list = [dict(zip(settings, t)) for t in it.product(*settings.values())]
    else:
        settings_list = [dict(zip(settings, t)) for t in zip(*settings.values())]

    if function:
        function = gf.get_cell(function)
        if not callable(function):
            raise ValueError(f"Error {function!r} needs to be callable.")
        component_list = [
            function(gf.get_component(doe, **settings)) for settings in settings_list
        ]
    else:
        component_list = [
            gf.get_component(doe, **settings) for settings in settings_list
        ]

    component_list = tuple(component_list)
    settings_list = tuple(settings_list)
    return component_list, settings_list


def pack_doe(
    doe: ComponentSpec = _doe,
    settings: dict[str, tuple[Any, ...]] = _settings,
    do_permutations: bool = False,
    function: CellSpec | None = None,
    **kwargs,
) -> Component:
    """Packs a component DOE (Design of Experiment) using pack.

    Args:
        doe: function to return Components.
        settings: component settings.
        do_permutations: for each setting.
        function: to apply (add padding, grating couplers).
        kwargs: for pack.

    Keyword Args:
        spacing: Minimum distance between adjacent shapes.
        aspect_ratio: (width, height) ratio of the rectangular bin.
        max_size: Limits the size into which the shapes will be packed.
        sort_by_area: Pre-sorts the shapes by area.
        density: Values closer to 1 pack tighter but require more computation.
        precision: Desired precision for rounding vertex coordinates.
        text: Optional function to add text labels.
        text_prefix: for labels. For example. 'A' for 'A1', 'A2'...
        text_offsets: relative to component size info anchor. Defaults to center.
        text_anchors: relative to component (ce cw nc ne nw sc se sw center cc).
        name_prefix: for each packed component (avoids the Unnamed cells warning).
            Note that the suffix contains a uuid so the name will not be deterministic.
        rotation: for each component in degrees.
        h_mirror: horizontal mirror in y axis (x, 1) (1, 0). This is the most common.
        v_mirror: vertical mirror using x axis (1, y) (0, y).
    """
    component_list, settings_list = generate_doe(
        doe, settings, do_permutations, function
    )

    c = pack(component_list, **kwargs)

    if len(c) > 1:
        raise ValueError(
            f"failed to pack in one Component, it created {len(c)} Components"
        )
    c = c[0]
    c.doe_names = [component.name for component in component_list]
    c.doe_settings = settings_list
    return c


def pack_doe_grid(
    doe: ComponentSpec = _doe,
    settings: dict[str, tuple[Any, ...]] = _settings,
    do_permutations: bool = False,
    function: CellSpec | None = None,
    with_text: bool = False,
    **kwargs,
) -> Component:
    """Packs a component DOE (Design of Experiment) using grid.

    Args:
        doe: function to return Components.
        settings: component settings.
        do_permutations: for each setting.
        function: to apply to component (add padding, grating couplers).
        with_text: includes text label.
        kwargs: for grid.

    Keyword Args:
        spacing: between adjacent elements on the grid, can be a tuple for
            different distances in height and width.
        separation: If True, guarantees elements are separated with fixed spacing
            if False, elements are spaced evenly along a grid.
        shape: x, y shape of the grid (see np.reshape).
            If no shape and the list is 1D, if np.reshape were run with (1, -1).
        align_x: {'x', 'xmin', 'xmax'} for x (column) alignment along.
        align_y: {'y', 'ymin', 'ymax'} for y (row) alignment along.
        edge_x: {'x', 'xmin', 'xmax'} for x (column) (ignored if separation = True).
        edge_y: {'y', 'ymin', 'ymax'} for y (row) (ignored if separation = True).
        rotation: for each component in degrees.
        h_mirror: horizontal mirror y axis (x, 1) (1, 0). most common mirror.
        v_mirror: vertical mirror using x axis (1, y) (0, y).
    """
    if do_permutations:
        settings_list = [dict(zip(settings, t)) for t in it.product(*settings.values())]
    else:
        settings_list = [dict(zip(settings, t)) for t in zip(*settings.values())]

    if function:
        function = gf.get_cell(function)
        if not callable(function):
            raise ValueError(f"Error {function!r} needs to be callable.")
        component_list = [
            function(gf.get_component(doe, **settings)) for settings in settings_list
        ]
    else:
        component_list = [
            gf.get_component(doe, **settings) for settings in settings_list
        ]

    component_list = tuple(component_list)
    settings_list = tuple(settings_list)

    if with_text:
        c = grid_with_text(component_list, **kwargs)

    else:
        c = grid(component_list, **kwargs)

    c.doe_names = [component.name for component in component_list]
    c.doe_settings = settings_list
    return c


if __name__ == "__main__":
    c = pack_doe_grid(
        doe="mmi1x2",
        settings=dict(length_mmi=(2.5, 100), width_mmi=(4, 10)),
        with_text=True,
        spacing=(100, 100),
        shape=(2, 2),
        do_permutations=True,
    )

    c = pack_doe(doe="mmi1x2", settings=dict(length_mmi=[2, 100], width_mmi=[4, 10]))
    # c = pack_doe()
    c.show()



"""Design of Experiment (DOE) with custom add_fiber_array function.

In this case add_fiber_array does not add labels.

You can use gf.add_labels.add_labels_to_ports.

"""

from __future__ import annotations

from functools import partial

import gdsfactory as gf

if __name__ == "__main__":
    c = gf.components.pack_doe(
        gf.components.straight,
        settings={"length": [5, 5]},
        function=partial(gf.routing.add_fiber_array, get_input_labels_function=None),
    )
    c = gf.add_labels.add_labels_to_ports(
        component=c, prefix="opt_te1550_", port_type="vertical_te"
    )
    print(len(c.labels))
    c.show()



from __future__ import annotations

from collections.abc import Iterable
from functools import partial
from inspect import getmembers, isfunction, signature

from kfactory import logger

from gdsfactory.typings import Any, Callable, Component


def get_cells(modules: Any, verbose: bool = False) -> dict[str, Callable]:
    """Returns PCells (component functions) from a module or list of modules.

    Args:
        modules: A module or an iterable of modules.
        verbose: If true, prints in case any errors occur.
    """
    # Ensure modules is iterable
    modules = modules if isinstance(modules, Iterable) else [modules]

    cells = {}
    for module in modules:
        for name, member in getmembers(module):
            if not name.startswith("_"):  # Exclude private names
                if callable(member) and (
                    isfunction(member) or isinstance(member, partial)
                ):
                    try:
                        r = signature(
                            member if not isinstance(member, partial) else member.func
                        ).return_annotation
                        if r == Component or (
                            isinstance(r, str) and r.endswith("Component")
                        ):
                            cells[name] = member
                    except ValueError as e:
                        if verbose:
                            logger.warn(f"error in {name}: {e}")
    return cells



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.typings import ComponentSpec


@gf.cell
def spiral(
    length: float = 100,
    bend: ComponentSpec = "bend_euler",
    straight: ComponentSpec = "straight",
    cross_section: ComponentSpec = "strip",
    spacing: float = 3.0,
    n_loops: int = 6,
) -> gf.Component:
    """Returns a spiral double (spiral in, and then out).

    Args:
        length: length of the spiral straight section.
        bend: bend component.
        straight: straight component.
        cross_section: cross_section component.
        spacing: spacing between the spiral loops.
        n_loops: number of loops.
    """
    c = gf.Component()
    xs = gf.get_cross_section(cross_section)

    b = gf.get_component(bend, cross_section=cross_section)
    radius = xs.radius
    _length = length

    b_inners = [c << b for _ in range(4)]
    b_inners[0].dmirror()
    b_inners[1].connect("o1", b_inners[0], "o2")
    b_inners[2].connect("o1", b_inners[1], "o2")
    s_space = c << gf.get_component(
        straight, cross_section=cross_section, length=spacing
    )
    s_space.connect("o1", b_inners[2], "o2")
    b_inners[3].connect("o1", s_space, "o2")
    l0_2 = c << gf.get_component(
        straight, cross_section=cross_section, length=_length + 2 * radius + spacing
    )
    l0_2.connect("o1", b_inners[3], "o2")
    p2 = l0_2.ports["o2"]

    if length > 0:
        l0_1 = c << gf.get_component(
            straight, cross_section=cross_section, length=_length
        )
        l0_1.connect("o1", b_inners[0], "o1")
        p1 = l0_1.ports["o2"].copy()
    else:
        p1 = b_inners[0].ports["o1"]
    p1.dmirror = not p1.dmirror
    for i in range(n_loops // 2):
        bends = [c << b for _ in range(8)]
        bends[0].connect("o1", p1)
        bends[1].connect("o1", p2)
        v1 = c << gf.get_component(
            straight, cross_section=cross_section, length=spacing * (1 + 4 * i)
        )
        v1.connect("o1", bends[0], "o2")
        v2 = c << gf.get_component(
            straight, cross_section=cross_section, length=spacing * (3 + 4 * i)
        )
        v2.connect("o1", bends[1], "o2")
        bends[2].connect("o1", v1, "o2")
        bends[3].connect("o1", v2, "o2")
        h1 = c << gf.get_component(
            straight,
            cross_section=cross_section,
            length=_length + 2 * radius + spacing * (1 + 4 * i),
        )
        h2 = c << gf.get_component(
            straight,
            cross_section=cross_section,
            length=_length + 2 * radius + spacing * (3 + 4 * i),
        )
        h1.connect("o1", bends[2], "o2")
        h2.connect("o1", bends[3], "o2")
        bends[4].connect("o1", h1, "o2")
        bends[5].connect("o1", h2, "o2")
        v3 = c << gf.get_component(
            straight, cross_section=cross_section, length=spacing * (3 + 4 * i)
        )
        v4 = c << gf.get_component(
            straight, cross_section=cross_section, length=spacing * (5 + 4 * i)
        )
        v3.connect("o1", bends[4], "o2")
        v4.connect("o1", bends[5], "o2")
        bends[6].connect("o1", v3, "o2")
        bends[7].connect("o1", v4, "o2")
        h3 = c << gf.get_component(
            straight,
            cross_section=cross_section,
            length=_length + 2 * radius + spacing * (3 + 4 * i),
        )
        h4 = c << gf.get_component(
            straight,
            cross_section=cross_section,
            length=_length + 2 * radius + spacing * (5 + 4 * i),
        )
        h3.connect("o1", bends[6], "o2")
        h4.connect("o1", bends[7], "o2")
        p1 = h3.ports["o2"]
        p2 = h4.ports["o2"]

    c.add_port(name="o1", port=p1)
    c.add_port(name="o2", port=p2)
    c.info["length"] = length * 2 * n_loops
    return c


if __name__ == "__main__":
    c = spiral(cross_section="rib", length=10, spacing=3.0)
    c.show()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import kfactory as kf
import numpy as np
from numpy import cos, float64, ndarray, sin

import gdsfactory as gf

RAD2DEG = 180.0 / np.pi
DEG2RAD = 1 / RAD2DEG

if TYPE_CHECKING:
    from gdsfactory.component import Component, Instance


def move_port_to_zero(
    component: Component, port_name: str = "o1", mirror: bool = False
) -> gf.Component:
    """Return a container that contains a reference to the original component.

    The new component has port_name in (0, 0).

    Args:
        component: to move the port to (0, 0).
        port_name: to move to (0, 0).
        mirror: if True, mirrors the component.
    """
    port_names = [p.name for p in component.ports]
    if port_name not in port_names:
        raise ValueError(f"port_name = {port_name!r} not in {port_names}")

    c = gf.Component()
    ref = c << component
    if mirror:
        ref.dmirror()

    movement = np.array(ref.ports[port_name].dcenter)
    ref.dmove(-movement)
    c.add_ports(ref.ports)
    c.copy_child_info(component)
    return c


def get_polygons(
    component_or_instance: Component | Instance,
    merge: bool = False,
    by: Literal["index"] | Literal["name"] | Literal["tuple"] = "index",
) -> dict[tuple[int, int] | str | int, list[kf.kdb.Polygon]]:
    """Returns a dict of Polygons per layer.

    Args:
        component_or_instance: to extract the polygons.
        merge: if True, merges the polygons.
        by: the format of the resulting keys in the dictionary ('index', 'name', 'tuple')
    """
    from gdsfactory import get_layer, get_layer_name

    if by == "index":
        get_key = get_layer
    elif by == "name":
        get_key = get_layer_name
    elif by == "tuple":

        def get_key(layer):
            return layer

    else:
        raise ValueError("argument 'by' should be 'index' | 'name' | 'tuple'")

    polygons = {}

    layers = [
        (info.layer, info.datatype)
        for info in component_or_instance.kcl.layer_infos()
        if not component_or_instance.bbox(component_or_instance.kcl.layer(info)).empty()
    ]
    c = (
        component_or_instance.parent_cell
        if hasattr(component_or_instance, "parent_cell")
        else component_or_instance
    )

    for layer in layers:
        layer_index = get_layer(layer)
        layer_key = get_key(layer)
        r = gf.kdb.Region(c.begin_shapes_rec(layer_index))
        if layer_key not in polygons:
            polygons[layer_key] = []
        if merge:
            r.merge()
        for p in r.each():
            polygons[layer_key].append(p)
    return polygons


def get_polygons_points(
    component_or_instance: Component | Instance,
    merge: bool = False,
    scale: float | None = None,
    by: Literal["index"] | Literal["name"] | Literal["tuple"] = "index",
) -> dict[int | str | tuple[int, int], list[tuple[float, float]]]:
    """Returns a dict with list of points per layer.

    Args:
        component_or_instance: to extract the polygons.
        merge: if True, merges the polygons.
        scale: if True, scales the points.
        by: the format of the resulting keys in the dictionary ('index', 'name', 'tuple')
    """
    polygons_dict = get_polygons(
        component_or_instance=component_or_instance, merge=merge, by=by
    )
    polygons_points = {}
    for layer, polygons in polygons_dict.items():
        all_points = []
        for polygon in polygons:
            if scale:
                points = np.array(
                    [
                        (point.x * scale, point.y * scale)
                        for point in polygon.to_simple_polygon()
                        .to_dtype(component_or_instance.kcl.dbu)
                        .each_point()
                    ]
                )
            else:
                points = np.array(
                    [
                        (point.x, point.y)
                        for point in polygon.to_simple_polygon()
                        .to_dtype(component_or_instance.kcl.dbu)
                        .each_point()
                    ]
                )
            all_points.append(points)
        polygons_points[layer] = all_points
    return polygons_points


def get_point_inside(component_or_instance: Component | Instance, layer) -> np.ndarray:
    """Returns a point inside the component or instance.

    Args:
        component_or_instance: to find a point inside.
        layer: to find a point inside.
    """
    layer = gf.get_layer(layer)
    return get_polygons_points(component_or_instance)[layer][0][0]


def sign_shape(pts: ndarray) -> float64:
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return np.sign((dx * y).sum())


def area(pts: ndarray) -> float64:
    """Returns the area."""
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return (dx * y).sum() / 2


def centered_diff(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - np.roll(a, 1, axis=0)) / 2
    return d[1:-1]


def centered_diff2(a: ndarray) -> ndarray:
    d = (np.roll(a, -1, axis=0) - a) - (a - np.roll(a, 1, axis=0))
    return d[1:-1]


def curvature(points: ndarray, t: ndarray) -> ndarray:
    """Args are the points and the tangents at each point.

        points : numpy.array shape (n, 2)
        t: numpy.array of size n

    Return:
        The curvature at each point.

    Computes the curvature at every point excluding the first and last point.

    For a planar curve parametrized as P(t) = (x(t), y(t)), the curvature is given
    by (x' y'' - x'' y' ) / (x' **2 + y' **2)**(3/2)

    """
    # Use centered difference for derivative
    dt = centered_diff(t)
    dp = centered_diff(points)
    dp2 = centered_diff2(points)

    dx = dp[:, 0] / dt
    dy = dp[:, 1] / dt

    dx2 = dp2[:, 0] / dt**2
    dy2 = dp2[:, 1] / dt**2

    return (dx * dy2 - dx2 * dy) / (dx**2 + dy**2) ** (3 / 2)


def radius_of_curvature(points, t):
    return 1 / curvature(points, t)


def path_length(points: ndarray) -> float64:
    """Returns: The path length.

    Args:
        points: With shape (N, 2) representing N points with coordinates x, y.
    """
    dpts = points[1:, :] - points[:-1, :]
    _d = dpts**2
    return np.sum(np.sqrt(_d[:, 0] + _d[:, 1]))


def snap_angle(a: float64) -> int:
    """Returns angle snapped along manhattan angle (0, 90, 180, 270).

    a: angle in deg
    Return angle snapped along manhattan angle
    """
    a = a % 360
    if -45 < a < 45:
        return 0
    elif 45 < a < 135:
        return 90
    elif 135 < a < 225:
        return 180
    elif 225 < a < 315:
        return 270
    else:
        return 0


def angles_rad(pts: ndarray) -> ndarray:
    """Returns the angles (radians) of the connection between each point and the next."""
    _pts = np.roll(pts, -1, 0)
    return np.arctan2(_pts[:, 1] - pts[:, 1], _pts[:, 0] - pts[:, 0])


def angles_deg(pts: ndarray) -> ndarray:
    """Returns the angles (degrees) of the connection between each point and the next."""
    return angles_rad(pts) * RAD2DEG


def extrude_path(
    points: ndarray,
    width: float,
    with_manhattan_facing_angles: bool = True,
    spike_length: float64 | int | float = 0,
    start_angle: int | None = None,
    end_angle: int | None = None,
    grid: float | None = None,
) -> ndarray:
    """Deprecated. Use gdsfactory.path.Path.extrude() instead.

    Extrude a path of `width` along a curve defined by `points`.

    Args:
        points: numpy 2D array of shape (N, 2).
        width: of the path to extrude.
        with_manhattan_facing_angles: snaps to manhattan angles.
        spike_length: in um.
        start_angle: in degrees.
        end_angle: in degrees.
        grid: in um.

    Returns:
        numpy 2D array of shape (2*N, 2).
    """
    grid = grid or gf.kcl.dbu

    if isinstance(points, list):
        points = np.stack([(p[0], p[1]) for p in points], axis=0)

    a = angles_deg(points)
    if with_manhattan_facing_angles:
        _start_angle = snap_angle(a[0] + 180)
        _end_angle = snap_angle(a[-2])
    else:
        _start_angle = a[0] + 180
        _end_angle = a[-2]

    start_angle = start_angle if start_angle is not None else _start_angle
    end_angle = end_angle if end_angle is not None else _end_angle

    a2 = angles_rad(points) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = end_angle * DEG2RAD - a2[-2]
    a1[0] = start_angle * DEG2RAD - a1[1]

    a_plus = a2 + a1
    cos_a_min = np.cos(a2 - a1)
    offsets = np.column_stack((-sin(a_plus) / cos_a_min, cos(a_plus) / cos_a_min)) * (
        0.5 * width
    )

    points_back = np.flipud(points - offsets)
    if spike_length != 0:
        d = spike_length
        a_start = start_angle * DEG2RAD
        a_end = end_angle * DEG2RAD
        p_start_spike = points[0] + d * np.array([[cos(a_start), sin(a_start)]])
        p_end_spike = points[-1] + d * np.array([[cos(a_end), sin(a_end)]])

        pts = np.vstack((p_start_spike, points + offsets, p_end_spike, points_back))
    else:
        pts = np.vstack((points + offsets, points_back))

    return np.round(pts / grid) * grid


if __name__ == "__main__":
    c = gf.c.rectangle(size=(10, 10), centered=True)
    # c = gf.Component()
    # ref = c << gf.components.mzi_lattice()
    # ref.dmovey(15)
    # p = get_polygons_points(ref)
    # p = get_point_inside(ref, layer=(1, 0))
    # c.add_label(text="hello", position=p)
    # c.show()
    c = move_port_to_zero(c, port_name="e4")
    c.show()



from gdsfactory.read.from_gdspaths import (
    from_gdsdir,
    from_gdspaths,
)
from gdsfactory.read.from_np import (
    from_image,
    from_np,
)
from gdsfactory.read.from_updk import (
    from_updk,
)
from gdsfactory.read.from_yaml import (
    from_yaml,
)
from gdsfactory.read.from_yaml_template import (
    cell_from_yaml_template,
)
from gdsfactory.read.import_gds import (
    import_gds,
)
from gdsfactory.read.labels import (
    add_port_markers,
    read_labels_yaml,
)

__all__ = [
    "add_port_markers",
    "cell_from_yaml_template",
    "from_gdsdir",
    "from_gdspaths",
    "from_image",
    "from_np",
    "from_updk",
    "from_yaml",
    "import_gds",
    "read_labels_yaml",
]



from __future__ import annotations

from collections import Counter

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec


class SequenceGenerator:
    def __init__(
        self,
        start_sequence: str = "IL",
        repeated_sequence: str = "ASASBSBS",
        end_sequence: str = "LO",
    ) -> None:
        """Sequence generator.

        Main use case: any type of cascade of components with repeating patterns
        such as serpentine, cutbacks etc...
        Component sequences have two ports by default.
        it adds aliases for the components forming the sequence.
        They use the component symbol with a suffix index starting from 1,
        so you may access the ports from any subcomponent.

        Usually we can break these components in 3 parts:
        - there is a starting pattern with input and possibly some special
        connections
        - then a repeating pattern
        - An ending pattern with an output

        Example of symbol meaning

        A: bend connected with input W0
        B: bend connected with input N0
        I: taper with input '1'
        O: taper with input '2'
        S: short straight waveguide
        L: long straight waveguide
        """
        self.start_sequence = start_sequence
        self.end_sequence = end_sequence
        self.repeated_sequence = repeated_sequence

    def get_sequence(self, n=2):
        return self.start_sequence + n * self.repeated_sequence + self.end_sequence


def parse_component_name(name: str) -> tuple[str, bool]:
    """If the component name has more than one character and starts with "!".

    then we need to flip along the axis given by the input port angle.
    """
    return (name[1:], True) if len(name) != 1 and name[0] == "!" else (name, False)


def _flip_ref(c_ref, port_name):
    a = c_ref.ports[port_name].orientation
    if a in [0, 180]:
        y = c_ref.ports[port_name].dcenter[1]
        c_ref.dmirror_y(y)
    else:
        c_ref.dmirror_x(port_name)
    return c_ref


def component_sequence(
    sequence: str,
    symbol_to_component: dict[str, tuple[ComponentSpec, str, str]],
    ports_map: dict[str, tuple[str, str]] | None = None,
    port_name1: str = "o1",
    port_name2: str = "o2",
    start_orientation: float = 0.0,
) -> Component:
    """Returns component from ASCII sequence.

    if you prefix a symbol with ! it mirrors the component

    Args:
        sequence: a string or a list of symbols.
        symbol_to_component: maps symbols to (component, input, output).
        ports_map: (optional) extra port mapping using the convention.
            {port_name: (alias_name, port_name)}
        port_name1: input port_name.
        port_name2: output port_name.
        start_orientation: in degrees.

    Returns:
        component: containing the sequence of sub-components
            instantiated and connected together in the sequence order.

    .. plot::
        :include-source:

        import gdsfactory as gf

        bend180 = gf.components.bend_circular180()
        wg_pin = gf.components.straight_pin(length=40)
        wg = gf.components.straight()

        # Define a map between symbols and (component, input port, output port)
        symbol_to_component = {
            "A": (bend180, 'o1', 'o2'),
            "B": (bend180, 'o2', 'o1'),
            "H": (wg_pin, 'o1', 'o2'),
            "-": (wg, 'o1', 'o2'),
        }

        # Each character in the sequence represents a component
        s = "AB-H-H-H-H-BA"
        c = gf.components.component_sequence(sequence=s, symbol_to_component=symbol_to_component)
        c.plot()
    """
    ports_map = ports_map or {}
    named_references_counter = Counter()
    component = Component()

    # Add first component reference and input port
    symbol = sequence[0] if "!" not in sequence[0] else sequence[:2]
    index = 2 if "!" in sequence[0] else 1
    name_start_device, do_flip = parse_component_name(symbol)
    component_input, input_port, prev_port = symbol_to_component[name_start_device]
    prev_device = component.add_ref(component_input, name=f"{symbol}{index}")
    named_references_counter.update(
        {name_start_device: 1}
    )  # sourcery skip:simplify-dictionary-update

    if do_flip:
        prev_device = _flip_ref(prev_device, input_port)

    prev_device.drotate(angle=start_orientation)

    try:
        component.add_port(name=port_name1, port=prev_device.ports[input_port])
    except KeyError as exc:
        port_names = [port.name for port in prev_device.ports]
        raise KeyError(
            f"{prev_device.parent.name!r} input_port {input_port!r} "
            f"not in {port_names}"
        ) from exc

    while index < len(sequence):
        s = sequence[index]

        if s == "!":
            # if it's the last character skip
            if index + 1 >= len(sequence):
                index += 1
                continue
            s = sequence[index + 1]
            do_flip = True
            index += 1
        else:
            do_flip = False

        index += 1
        component_i, input_port, next_port = symbol_to_component[s]
        component_i = gf.get_component(component_i)
        named_references_counter.update(
            {s: 1}
        )  # sourcery skip:simplify-dictionary-update
        alias = f"{s}{named_references_counter[s]}"
        ref = component.add_ref(component_i, name=alias)

        if do_flip:
            ref = _flip_ref(ref, input_port)

        try:
            ref.connect(input_port, prev_device.ports[prev_port])
        except KeyError as exc:
            port_names = [port.name for port in prev_device.ports]
            raise KeyError(
                f"{prev_device.parent.name!r} port {prev_port!r} "
                f"not in {port_names}"
            ) from exc

        prev_device = ref
        prev_port = next_port

    # Deal with edge case where the sequence contains only one component
    if len(sequence) == 1:
        ref = prev_device
        next_port = prev_port

    try:
        component.add_port(name=port_name2, port=ref.ports[next_port])
    except BaseException:
        print(sequence)
        raise

    # Add any extra port specified in ports_map
    for name, (ref_name, alias_port_name) in ports_map.items():
        component.add_port(
            name=name, port=component.insts[ref_name].ports[alias_port_name]
        )

    return component


if __name__ == "__main__":
    import gdsfactory as gf

    bend180 = gf.components.bend_circular180()
    wg_pin = gf.components.straight_pin(length=40)
    wg = gf.components.straight()

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component_map = {
        "A": (bend180, "o1", "o2"),
        "B": (bend180, "o2", "o1"),
        "H": (wg_pin, "o1", "o2"),
        "-": (wg, "o1", "o2"),
    }

    # Each character in the sequence represents a component
    sequence = "AB"
    sequence = "HH!"
    sequence = "!HH"
    sequence = "H"
    sequence = "AB-H-H-H-H-BA"
    c = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component_map
    )
    c.show()
    # n = c.get_netlist()
    # c = gf.read.from_yaml(n)

    # _ = c << gf.c.straight()
    # c.name = "top"
    # lyrdb = c.connectivity_check()
    # gf.show(c, lyrdb=lyrdb)



"""Each component factory component returns a component.

Make sure your components get imported here so the PDK registers them.
"""

from __future__ import annotations

import sys
from functools import partial

from gdsfactory.components.add_fiber_array_optical_south_electrical_north import (
    add_fiber_array_optical_south_electrical_north,
)
from gdsfactory.components.add_termination import add_termination
from gdsfactory.components.add_trenches import add_trenches, add_trenches90
from gdsfactory.components.align import add_frame, align_wafer
from gdsfactory.components.array_component import array
from gdsfactory.components.bbox import bbox
from gdsfactory.components.bend_circular import (
    bend_circular,
    bend_circular180,
    bend_circular_all_angle,
)
from gdsfactory.components.bend_circular_heater import bend_circular_heater
from gdsfactory.components.bend_euler import (
    bend_euler,
    bend_euler180,
    bend_euler_all_angle,
    bend_euler_s,
)
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.bezier import bezier
from gdsfactory.components.C import C
from gdsfactory.components.cavity import cavity
from gdsfactory.components.cdsem_all import cdsem_all
from gdsfactory.components.cdsem_bend180 import cdsem_bend180
from gdsfactory.components.cdsem_coupler import cdsem_coupler
from gdsfactory.components.cdsem_straight import cdsem_straight
from gdsfactory.components.cdsem_straight_density import cdsem_straight_density
from gdsfactory.components.circle import circle
from gdsfactory.components.coh_rx_single_pol import coh_rx_single_pol
from gdsfactory.components.coh_tx_dual_pol import coh_tx_dual_pol
from gdsfactory.components.coh_tx_single_pol import coh_tx_single_pol
from gdsfactory.components.compass import compass
from gdsfactory.components.component_sequence import component_sequence
from gdsfactory.components.copy_layers import copy_layers
from gdsfactory.components.coupler import coupler
from gdsfactory.components.coupler90 import coupler90, coupler90circular
from gdsfactory.components.coupler90bend import coupler90bend
from gdsfactory.components.coupler_adiabatic import coupler_adiabatic
from gdsfactory.components.coupler_asymmetric import coupler_asymmetric
from gdsfactory.components.coupler_bent import coupler_bent
from gdsfactory.components.coupler_broadband import coupler_broadband
from gdsfactory.components.coupler_full import coupler_full
from gdsfactory.components.coupler_ring import coupler_ring
from gdsfactory.components.coupler_straight import coupler_straight
from gdsfactory.components.coupler_straight_asymmetric import (
    coupler_straight_asymmetric,
)
from gdsfactory.components.coupler_symmetric import coupler_symmetric
from gdsfactory.components.cross import cross
from gdsfactory.components.crossing_waveguide import (
    crossing,
    crossing45,
    crossing_arm,
    crossing_etched,
    crossing_from_taper,
)
from gdsfactory.components.cutback_2x2 import cutback_2x2
from gdsfactory.components.cutback_bend import (
    cutback_bend,
    cutback_bend90,
    cutback_bend90circular,
    cutback_bend180,
    cutback_bend180circular,
    staircase,
)
from gdsfactory.components.cutback_component import (
    cutback_component,
    cutback_component_mirror,
)
from gdsfactory.components.cutback_loss import (
    cutback_loss,
    cutback_loss_bend90,
    cutback_loss_bend180,
    cutback_loss_mmi1x2,
    cutback_loss_spirals,
)
from gdsfactory.components.cutback_splitter import cutback_splitter
from gdsfactory.components.dbr import dbr
from gdsfactory.components.dbr_tapered import dbr_tapered
from gdsfactory.components.delay_snake import delay_snake
from gdsfactory.components.delay_snake2 import delay_snake2
from gdsfactory.components.delay_snake_sbend import delay_snake_sbend
from gdsfactory.components.dicing_lane import dicing_lane
from gdsfactory.components.die import die
from gdsfactory.components.die_bbox import die_bbox
from gdsfactory.components.die_with_pads import die_with_pads
from gdsfactory.components.disk import disk, disk_heater
from gdsfactory.components.edge_coupler_array import (
    edge_coupler_array,
    edge_coupler_array_with_loopback,
    edge_coupler_silicon,
)
from gdsfactory.components.ellipse import ellipse
from gdsfactory.components.extend_ports_list import extend_ports_list
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.fiber import fiber
from gdsfactory.components.fiber_array import fiber_array
from gdsfactory.components.fiducial_squares import fiducial_squares
from gdsfactory.components.ge_detector_straight_si_contacts import (
    ge_detector_straight_si_contacts,
)
from gdsfactory.components.grating_coupler_array import grating_coupler_array
from gdsfactory.components.grating_coupler_dual_pol import grating_coupler_dual_pol
from gdsfactory.components.grating_coupler_elliptical import (
    ellipse_arc,
    grating_coupler_elliptical,
    grating_coupler_elliptical_te,
    grating_coupler_elliptical_tm,
    grating_taper_points,
    grating_tooth_points,
)
from gdsfactory.components.grating_coupler_elliptical_arbitrary import (
    grating_coupler_elliptical_arbitrary,
    grating_coupler_elliptical_uniform,
)
from gdsfactory.components.grating_coupler_elliptical_lumerical import (
    grating_coupler_elliptical_lumerical,
)
from gdsfactory.components.grating_coupler_elliptical_trenches import (
    grating_coupler_elliptical_trenches,
    grating_coupler_te,
    grating_coupler_tm,
)
from gdsfactory.components.grating_coupler_loss import (
    grating_coupler_loss_fiber_array,
    grating_coupler_loss_fiber_array4,
    loss_deembedding_ch12_34,
    loss_deembedding_ch13_24,
    loss_deembedding_ch14_23,
)
from gdsfactory.components.grating_coupler_rectangular import (
    grating_coupler_rectangular,
)
from gdsfactory.components.grating_coupler_rectangular_arbitrary import (
    grating_coupler_rectangular_arbitrary,
)
from gdsfactory.components.grating_coupler_tree import grating_coupler_tree
from gdsfactory.components.greek_cross import (
    greek_cross,
    greek_cross_with_pads,
)
from gdsfactory.components.hline import hline
from gdsfactory.components.interdigital_capacitor import interdigital_capacitor
from gdsfactory.components.L import L
from gdsfactory.components.litho_calipers import litho_calipers
from gdsfactory.components.litho_ruler import litho_ruler
from gdsfactory.components.litho_steps import litho_steps
from gdsfactory.components.loop_mirror import loop_mirror
from gdsfactory.components.mmi import mmi
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.mmi1x2_with_sbend import mmi1x2_with_sbend
from gdsfactory.components.mmi2x2 import mmi2x2
from gdsfactory.components.mmi2x2_with_sbend import mmi2x2_with_sbend
from gdsfactory.components.mmi_90degree_hybrid import mmi_90degree_hybrid
from gdsfactory.components.mode_converter import mode_converter
from gdsfactory.components.mzi import (
    mzi,
    mzi1x2_2x2,
    mzi2x2_2x2,
    mzi2x2_2x2_phase_shifter,
    mzi_coupler,
    mzi_phase_shifter,
    mzi_phase_shifter_top_heater_metal,
    mzi_pin,
    mzm,
)
from gdsfactory.components.mzi_arm import mzi_arm
from gdsfactory.components.mzi_arms import mzi_arms
from gdsfactory.components.mzi_lattice import mzi_lattice, mzi_lattice_mmi
from gdsfactory.components.mzi_pads_center import mzi_pads_center
from gdsfactory.components.mzit import mzit
from gdsfactory.components.mzit_lattice import mzit_lattice
from gdsfactory.components.nxn import nxn
from gdsfactory.components.optimal_90deg import optimal_90deg
from gdsfactory.components.optimal_hairpin import optimal_hairpin
from gdsfactory.components.optimal_step import optimal_step
from gdsfactory.components.pack_doe import generate_doe, pack_doe, pack_doe_grid
from gdsfactory.components.pad import (
    pad,
    pad_array,
    pad_array0,
    pad_array90,
    pad_array180,
    pad_array270,
    pad_rectangular,
    pad_small,
)
from gdsfactory.components.pad_gsg import pad_gsg_open, pad_gsg_short
from gdsfactory.components.pads_shorted import pads_shorted
from gdsfactory.components.polarization_splitter_rotator import (
    polarization_splitter_rotator,
)
from gdsfactory.components.ramp import ramp
from gdsfactory.components.rectangle import rectangle, rectangles
from gdsfactory.components.rectangle_with_slits import rectangle_with_slits
from gdsfactory.components.regular_polygon import hexagon, octagon, regular_polygon
from gdsfactory.components.resistance_meander import resistance_meander
from gdsfactory.components.resistance_sheet import resistance_sheet
from gdsfactory.components.ring import ring
from gdsfactory.components.ring_crow import ring_crow
from gdsfactory.components.ring_crow_couplers import ring_crow_couplers
from gdsfactory.components.ring_double import ring_double
from gdsfactory.components.ring_double_pn import ring_double_pn
from gdsfactory.components.ring_heater import ring_double_heater, ring_single_heater
from gdsfactory.components.ring_single import ring_single
from gdsfactory.components.ring_single_array import ring_single_array
from gdsfactory.components.ring_single_bend_coupler import (
    coupler_bend,
    ring_single_bend_coupler,
)
from gdsfactory.components.ring_single_dut import ring_single_dut, taper2
from gdsfactory.components.ring_single_pn import ring_single_pn
from gdsfactory.components.seal_ring import seal_ring, seal_ring_segmented
from gdsfactory.components.snspd import snspd
from gdsfactory.components.spiral import spiral
from gdsfactory.components.spiral_double import spiral_double
from gdsfactory.components.spiral_heater import (
    spiral_racetrack,
    spiral_racetrack_fixed_length,
    spiral_racetrack_heater_doped,
    spiral_racetrack_heater_metal,
)
from gdsfactory.components.spiral_inductor import spiral_inductor
from gdsfactory.components.splitter_chain import splitter_chain
from gdsfactory.components.splitter_tree import splitter_tree, switch_tree
from gdsfactory.components.straight import straight, straight_all_angle
from gdsfactory.components.straight_array import straight_array
from gdsfactory.components.straight_heater_doped import (
    straight_heater_doped_rib,
    straight_heater_doped_strip,
)
from gdsfactory.components.straight_heater_meander import straight_heater_meander
from gdsfactory.components.straight_heater_meander_doped import (
    straight_heater_meander_doped,
)
from gdsfactory.components.straight_heater_metal import (
    straight_heater_metal,
    straight_heater_metal_90_90,
    straight_heater_metal_simple,
    straight_heater_metal_undercut,
    straight_heater_metal_undercut_90_90,
)
from gdsfactory.components.straight_pin import straight_pin, straight_pn
from gdsfactory.components.straight_pin_slot import straight_pin_slot
from gdsfactory.components.taper import (
    taper,
    taper_sc_nc,
    taper_strip_to_ridge,
    taper_strip_to_ridge_trenches,
)
from gdsfactory.components.taper_adiabatic import taper_adiabatic
from gdsfactory.components.taper_cross_section import (
    taper_cross_section,
    taper_cross_section_linear,
    taper_cross_section_parabolic,
    taper_cross_section_sine,
)
from gdsfactory.components.taper_from_csv import taper_from_csv
from gdsfactory.components.taper_parabolic import taper_parabolic
from gdsfactory.components.terminator import terminator
from gdsfactory.components.text import text, text_klayout, text_lines
from gdsfactory.components.text_freetype import text_freetype
from gdsfactory.components.text_rectangular import (
    text_rectangular,
    text_rectangular_mini,
    text_rectangular_multi_layer,
)
from gdsfactory.components.triangles import triangle, triangle2, triangle4
from gdsfactory.components.verniers import verniers
from gdsfactory.components.version_stamp import pixel, qrcode, version_stamp
from gdsfactory.components.via import via, via1, via2, viac
from gdsfactory.components.via_chain import via_chain
from gdsfactory.components.via_corner import via_corner
from gdsfactory.components.via_stack import (
    via_stack,
    via_stack_corner45,
    via_stack_corner45_extended,
    via_stack_heater_m3,
    via_stack_heater_mtop,
    via_stack_heater_mtop_mini,
    via_stack_m1_mtop,
    via_stack_npp_m1,
    via_stack_slab_m1_horizontal,
    via_stack_slab_m3,
    via_stack_slab_npp_m3,
)
from gdsfactory.components.via_stack_with_offset import via_stack_with_offset
from gdsfactory.components.wafer import wafer
from gdsfactory.components.wire import wire_corner, wire_corner45, wire_straight
from gdsfactory.get_factories import get_cells

bend_euler_trenches = partial(add_trenches90, component=bend_euler)
coupler_trenches = partial(add_trenches, component=coupler)
ring_single_trenches = partial(add_trenches, component=ring_single)
ring_double_trenches = partial(add_trenches, component=ring_double)


__all__ = [
    "add_termination",
    "C",
    "L",
    "add_fiber_array_optical_south_electrical_north",
    "add_frame",
    "add_trenches",
    "align_wafer",
    "array",
    "bbox",
    "bend_circular",
    "bend_circular180",
    "bend_circular_heater",
    "bend_circular_all_angle",
    "bend_euler",
    "bend_euler_all_angle",
    "bend_euler180",
    "bend_euler_s",
    "bend_s",
    "bezier",
    "cavity",
    "cdsem_all",
    "cdsem_straight",
    "cdsem_straight_density",
    "cdsem_coupler",
    "cdsem_bend180",
    "circle",
    "coh_rx_single_pol",
    "coh_tx_dual_pol",
    "coh_tx_single_pol",
    "compass",
    "component_sequence",
    "copy_layers",
    "coupler",
    "coupler90",
    "coupler90bend",
    "coupler90circular",
    "coupler_adiabatic",
    "coupler_asymmetric",
    "coupler_bend",
    "coupler_bent",
    "coupler_broadband",
    "coupler_full",
    "coupler_ring",
    "coupler_straight",
    "coupler_straight_asymmetric",
    "coupler_symmetric",
    "coupler_trenches",
    "cross",
    "crossing",
    "crossing45",
    "crossing_arm",
    "crossing_etched",
    "crossing_from_taper",
    "cutback_2x2",
    "cutback_bend",
    "cutback_bend180",
    "cutback_bend180circular",
    "cutback_bend90",
    "cutback_bend90circular",
    "cutback_component",
    "cutback_component_mirror",
    "cutback_loss",
    "cutback_loss_bend90",
    "cutback_loss_bend180",
    "cutback_loss_mmi1x2",
    "cutback_loss_spirals",
    "cutback_splitter",
    "dbr",
    "dbr_tapered",
    "delay_snake",
    "delay_snake2",
    "delay_snake_sbend",
    "dicing_lane",
    "die",
    "die_bbox",
    "die_with_pads",
    "disk",
    "disk_heater",
    "edge_coupler_array",
    "edge_coupler_array_with_loopback",
    "edge_coupler_silicon",
    "ellipse",
    "ellipse_arc",
    "extend_ports",
    "extend_ports_list",
    "fiber",
    "fiber_array",
    "fiducial_squares",
    "ge_detector_straight_si_contacts",
    "generate_doe",
    "grating_coupler_array",
    "grating_coupler_dual_pol",
    "grating_coupler_elliptical",
    "grating_coupler_elliptical_arbitrary",
    "grating_coupler_elliptical_lumerical",
    "grating_coupler_elliptical_te",
    "grating_coupler_elliptical_tm",
    "grating_coupler_elliptical_trenches",
    "grating_coupler_elliptical_uniform",
    "grating_coupler_loss_fiber_array",
    "grating_coupler_loss_fiber_array4",
    "grating_coupler_rectangular",
    "grating_coupler_rectangular_arbitrary",
    "grating_coupler_te",
    "grating_coupler_tm",
    "grating_coupler_tree",
    "grating_taper_points",
    "grating_tooth_points",
    "greek_cross",
    "greek_cross_with_pads",
    "hline",
    "interdigital_capacitor",
    "litho_calipers",
    "litho_ruler",
    "litho_steps",
    "loop_mirror",
    "loss_deembedding_ch12_34",
    "loss_deembedding_ch13_24",
    "loss_deembedding_ch14_23",
    "mmi",
    "mmi1x2",
    "mmi1x2_with_sbend",
    "mmi2x2",
    "mmi2x2_with_sbend",
    "mmi_90degree_hybrid",
    "mode_converter",
    "mzi",
    "mzi1x2_2x2",
    "mzi2x2_2x2",
    "mzi2x2_2x2_phase_shifter",
    "mzi_arm",
    "mzi_arms",
    "mzi_coupler",
    "mzi_lattice",
    "mzi_lattice_mmi",
    "mzi_pads_center",
    "mzi_phase_shifter",
    "mzi_phase_shifter_top_heater_metal",
    "mzi_pin",
    "mzit",
    "mzit_lattice",
    "mzm",
    "nxn",
    "optimal_90deg",
    "optimal_hairpin",
    "optimal_step",
    "pack_doe",
    "pack_doe_grid",
    "pad",
    "pad_array",
    "pad_array0",
    "pad_array180",
    "pad_array270",
    "pad_array90",
    "pad_gsg_open",
    "pad_gsg_short",
    "pad_rectangular",
    "pad_small",
    "pads_shorted",
    "pixel",
    "polarization_splitter_rotator",
    "qrcode",
    "ramp",
    "rectangle",
    "rectangle_with_slits",
    "rectangles",
    "regular_polygon",
    "resistance_meander",
    "resistance_sheet",
    "ring",
    "ring_crow",
    "ring_crow_couplers",
    "ring_double",
    "ring_double_heater",
    "ring_double_pn",
    "ring_double_trenches",
    "ring_single",
    "ring_single_array",
    "ring_single_bend_coupler",
    "ring_single_dut",
    "ring_single_heater",
    "ring_single_pn",
    "ring_single_trenches",
    "seal_ring",
    "seal_ring_segmented",
    "snspd",
    "spiral",
    "spiral_inductor",
    "spiral_double",
    "spiral_racetrack",
    "spiral_racetrack_fixed_length",
    "spiral_racetrack_heater_doped",
    "spiral_racetrack_heater_metal",
    "splitter_chain",
    "splitter_tree",
    "staircase",
    "straight",
    "straight_all_angle",
    "straight_array",
    "straight_heater_doped_rib",
    "straight_heater_doped_strip",
    "straight_heater_meander",
    "straight_heater_meander_doped",
    "straight_heater_metal",
    "straight_heater_metal_90_90",
    "straight_heater_metal_simple",
    "straight_heater_metal_undercut",
    "straight_heater_metal_undercut_90_90",
    "straight_pin",
    "straight_pin_slot",
    "straight_pn",
    "switch_tree",
    "taper",
    "taper2",
    "taper_adiabatic",
    "taper_cross_section",
    "taper_cross_section_linear",
    "taper_cross_section_parabolic",
    "taper_cross_section_sine",
    "taper_from_csv",
    "taper_parabolic",
    "taper_sc_nc",
    "taper_strip_to_ridge",
    "taper_strip_to_ridge_trenches",
    "terminator",
    "text",
    "text_klayout",
    "text_freetype",
    "text_lines",
    "text_rectangular",
    "text_rectangular_multi_layer",
    "text_rectangular_mini",
    "triangle",
    "triangle2",
    "triangle4",
    "verniers",
    "version_stamp",
    "via",
    "via1",
    "via2",
    "via_chain",
    "via_corner",
    "via_stack",
    "via_stack_heater_m3",
    "via_stack_m1_mtop",
    "via_stack_slab_m1_horizontal",
    "via_stack_heater_mtop",
    "via_stack_heater_mtop_mini",
    "via_stack_slab_m3",
    "via_stack_slab_npp_m3",
    "via_stack_with_offset",
    "via_stack_npp_m1",
    "via_stack_corner45",
    "via_stack_corner45_extended",
    "viac",
    "wafer",
    "wire_corner",
    "wire_corner45",
    "wire_straight",
    "hexagon",
    "octagon",
]

cells = get_cells(sys.modules[__name__])



from __future__ import annotations

import gdsfactory as gf
from gdsfactory import cell
from gdsfactory.component import Component
from gdsfactory.components.dbr import dbr
from gdsfactory.typings import ComponentSpec


@cell
def cavity(
    component: ComponentSpec = dbr,
    coupler: ComponentSpec = "coupler",
    length: float = 0.1,
    gap: float = 0.2,
    **kwargs,
) -> Component:
    r"""Returns  cavity from a coupler and a mirror.

    connects the W0 port of the mirror to E1 and W1 coupler ports
    creating a resonant cavity

    Args:
        component: mirror.
        coupler: coupler library.
        length: coupler length.
        gap: coupler gap.
        kwargs: coupler_settings.

    .. code::

      ml (mirror left)              mr (mirror right)
       |                               |
       |o1 - o2__             __o3 - o1|
       |         \           /         |
                  \         /
                ---=========---
         o1  o1    length      o4    o2

    """
    mirror = gf.get_component(component)
    coupler = gf.get_component(coupler, length=length, gap=gap, **kwargs)

    c = gf.Component()
    c.component = mirror
    cr = c << coupler
    ml = c << mirror
    mr = c << mirror

    ml.connect("o1", other=cr.ports["o2"])
    mr.connect("o1", other=cr.ports["o3"])
    c.add_port("o1", port=cr.ports["o1"])
    c.add_port("o2", port=cr.ports["o4"])
    c.copy_child_info(mirror)
    return c


if __name__ == "__main__":
    from gdsfactory.components.dbr import dbr

    c = cavity(component=dbr())
    c.show()



from __future__ import annotations

from collections.abc import Callable

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.pad import pad_rectangular
from gdsfactory.components.straight_heater_metal import straight_heater_metal
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing.route_fiber_array import route_fiber_array
from gdsfactory.typings import (
    ComponentSpec,
    CrossSectionSpec,
    Strs,
)


def add_pads_bot(
    component: ComponentSpec = straight_heater_metal,
    select_ports: Callable = select_ports_electrical,
    port_names: Strs | None = None,
    component_name: str | None = None,
    cross_section: CrossSectionSpec = "metal_routing",
    pad_port_name: str = "e1",
    pad: ComponentSpec = pad_rectangular,
    bend: ComponentSpec = "wire_corner",
    straight_separation: float = 15.0,
    pad_spacing: float | str = "pad_spacing",
    optical_routing_type: int | None = 1,
    taper: ComponentSpec | None = None,
    port_type: str = "electrical",
    allow_width_mismatch: bool = True,
    fanout_length: float | None = 80,
    route_width: float | list[float] | None = 0,
    **kwargs,
) -> Component:
    """Returns new component with ports connected bottom pads.

    Args:
        component: component spec to connect to.
        select_ports: function to select_ports.
        port_names: optional port names. Overrides select_ports.
        component_name: optional for the label.
        cross_section: cross_section spec.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        pad_port_name: pad input port name.
        pad_port_labels: pad list of labels.
        pad: spec for route terminations.
        bend: bend spec.
        straight_separation: from wire edge to edge. Defaults to xs.width+xs.gap
        pad_spacing: in um. Defaults to pad_spacing constant from the PDK.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.
        taper: taper spec.
        port_type: port type.
        allow_width_mismatch: True
        fanout_length: if None, automatic calculation of fanout length.
        route_width: width of the route. If None, defaults to cross_section.width.
        kwargs: additional arguments.

    Keyword Args:
        straight: straight spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: route_single.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.
        allow_width_mismatch: True

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.c.nxn(
            xsize=600,
            ysize=200,
            north=2,
            south=3,
            wg_width=10,
            layer="M3",
            port_type="electrical",
        )
        cc = gf.routing.add_pads_bot(component=c, port_names=("e1", "e4"), fanout_length=50)
        cc.plot()

    """
    component_new = Component()
    component = gf.get_component(component)
    component_name = component_name or component.name

    pad_spacing = gf.get_constant(pad_spacing)
    cref = component_new << component
    ports = [cref[port_name] for port_name in port_names] if port_names else None
    ports = ports or select_ports(cref.ports)

    pad_component = gf.get_component(pad)
    if pad_port_name not in pad_component.ports:
        pad_ports = list(pad_component.ports)
        raise ValueError(
            f"pad_port_name = {pad_port_name!r} not in {pad_component.name!r} ports {pad_ports}"
        )

    pad_orientation = int(pad_component[pad_port_name].orientation)
    if pad_orientation != 180:
        raise ValueError(
            f"port.orientation={pad_orientation} for port {pad_port_name!r} needs to be 180 degrees."
        )

    if not ports:
        raise ValueError(
            f"select_ports or port_names did not match any ports in {list(component.ports)}"
        )

    route_fiber_array(
        component_new,
        component,
        grating_coupler=pad,
        gc_port_name=pad_port_name,
        component_name=component_name,
        cross_section=cross_section,
        select_ports=select_ports,
        with_loopback=False,
        bend=bend,
        straight_separation=straight_separation,
        port_names=port_names,
        fiber_spacing=pad_spacing,
        optical_routing_type=optical_routing_type,
        taper=taper,
        port_type=port_type,
        gc_port_name_fiber=pad_port_name,
        allow_width_mismatch=allow_width_mismatch,
        fanout_length=fanout_length,
        route_width=route_width,
        **kwargs,
    )
    component_new.add_ref(component)
    component_new.copy_child_info(component)
    return component_new


def add_pads_top(
    component: ComponentSpec = straight_heater_metal, **kwargs
) -> Component:
    """Returns new component with ports connected top pads.

    Args:
        component: component spec to connect to.
        kwargs: additional arguments.

    Keyword Args:
        select_ports: function to select_ports.
        port_names: optional port names. Overrides select_ports.
        component_name: optional for the label.
        cross_section: cross_section function.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        pad_port_name: pad input port name.
        pad_port_labels: pad list of labels.
        pad: spec for route terminations.
        bend: bend spec.
        straight_separation: from edge to edge.
        straight: straight spec.
        taper: taper spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        fanout_length: if None, automatic calculation of fanout length.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: route_single.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.c.nxn(
            xsize=600,
            ysize=200,
            north=2,
            south=3,
            wg_width=10,
            layer="M3",
            port_type="electrical",
        )
        cc = gf.routing.add_pads_top(component=c, port_names=("e1", "e4"), fanout_length=50)
        cc.plot()

    """
    c = Component()
    _c = add_pads_bot(component=component, **kwargs)
    ref = c << _c
    ref.mirror_y()
    c.add_ports(ref.ports)
    c.copy_child_info(_c)
    return c


if __name__ == "__main__":
    # c = gf.components.pad()
    c = gf.components.straight_heater_metal(length=100.0)
    # c = gf.components.straight(length=100.0)
    c = gf.routing.add_pads_top(component=c, port_names=("l_e1",))
    c = gf.routing.add_fiber_array(c)
    c.show()
    # c.show()

    # cc = add_pads_top(component=c, port_names=("e1",))
    # cc = add_pads_top(component=c, port_names=("e1", "e2"), fanout_length=50)
    # c = gf.c.nxn(
    #     xsize=600,
    #     ysize=200,
    #     # north=2,
    #     # south=3,
    #     north=0,
    #     south=0,
    #     west=2,
    #     east=2,
    #     wg_width=10,
    #     layer="M3",
    #     port_type="electrical",
    # )
    # cc = add_pads_top(component=c)
    # cc.show()



"""Read component from a numpy.ndarray."""

from __future__ import annotations

import numpy as np

from gdsfactory import cell
from gdsfactory.boolean import boolean
from gdsfactory.component import Component


def compute_area_signed(pr) -> float:
    """Return the signed area enclosed by a ring using the linear time.

    algorithm at http://www.cgafaq.info/wiki/Polygon_Area. A value >= 0
    indicates a counter-clockwise oriented ring.

    """
    xs, ys = map(list, zip(*pr))
    xs.append(xs[1])
    ys.append(ys[1])
    return sum(xs[i] * (ys[i + 1] - ys[i - 1]) for i in range(1, len(pr))) / 2.0


def from_np(
    ndarray,
    nm_per_pixel: int = 20,
    layer: tuple[int, int] = (1, 0),
    threshold: float = 0.99,
    invert: bool = True,
) -> Component:
    """Returns Component from a np.ndarray.

    Extracts contours skimage.measure.find_contours using `threshold`.

    Args:
        ndarray: 2D ndarray representing the device layout.
        nm_per_pixel: scale_factor.
        layer: layer tuple to output gds.
        threshold: value along which to find contours in the array.
        invert: invert the mask.
    """
    from skimage import measure

    c = Component()
    d = Component()
    ndarray = np.pad(ndarray, 2)
    contours = measure.find_contours(ndarray, threshold)
    assert len(contours) > 0, (
        f"no contours found for threshold = {threshold}, maybe you can reduce the"
        " threshold"
    )

    for contour in contours:
        area = compute_area_signed(contour)
        points = contour * 1e-3 * nm_per_pixel
        if area < 0:
            c.add_polygon(points, layer=layer)
        else:
            d.add_polygon(points, layer=layer)

    return boolean(c, d, operation="not", layer=layer) if invert else d


@cell
def from_image(image_path: str, **kwargs) -> Component:
    """Returns Component from a png image.

    Args:
        image_path: png file path.
        kwargs: for from_np.

    Keyword Args:
        nm_per_pixel: scale_factor.
        layer: layer tuple to output gds.
        threshold: value along which to find contours in the array.

    """
    import matplotlib.pyplot as plt

    # Load the image using matplotlib
    img = plt.imread(image_path)

    if len(img.shape) == 3:
        img = 0.2989 * img[:, :, 0] + 0.5870 * img[:, :, 1] + 0.1140 * img[:, :, 2]

    # Convert image to numpy array (in fact, plt.imread already returns a numpy array)
    img_array = np.array(img)

    return from_np(img_array, **kwargs)


if __name__ == "__main__":
    from gdsfactory.config import PATH

    # import gdsfactory as gf
    # c1 = gf.components.straight()
    # c1 = gf.components.bend_circular()
    # c1 = gf.components.ring_single()
    # img = c1.to_np()
    # c2 = from_np(img)
    # c2.show()

    c = from_image(
        PATH.module / "samples" / "images" / "logo.png", nm_per_pixel=500, invert=True
    )
    c.show()



from gdsfactory.labels.add_label_yaml import (
    add_label_json,
    add_label_yaml,
)
from gdsfactory.labels.ehva import (
    add_label_ehva,
    ignore,
    prefix_to_type_default,
)
from gdsfactory.labels.get_test_manifest import get_test_manifest
from gdsfactory.labels.write_labels import (
    find_labels,
    write_labels,
)
from gdsfactory.labels.write_test_manifest import (
    write_test_manifest,
)

__all__ = [
    "add_label_ehva",
    "add_label_json",
    "add_label_yaml",
    "find_labels",
    "get_test_manifest",
    "ignore",
    "prefix_to_type_default",
    "write_labels",
    "write_test_manifest",
]



from gdsfactory.typings import Layer, PathType


def extract_layers(input_text: str) -> dict[str, Layer]:
    # Split the input text by lines
    lines = input_text.split("\n")

    # A dictionary to store the mapping of the layer name to its tuple
    layer_mapping = {}

    # For each line in the text
    for line in lines:
        # Split the line by spaces
        parts = line.split()
        if line.startswith("#"):
            continue

        # Check if the line has the expected number of parts (columns)
        if len(parts) == 7:  # Assuming your data structure remains consistent
            layer_name = parts[0]
            layer_num = parts[4]
            data_type = parts[5]

            # Convert layer number and data type to integers
            try:
                layer_num = int(layer_num)
                data_type = int(data_type)

                # Store the layer name mapping
                layer_mapping[layer_name] = (layer_num, data_type)

            except (
                ValueError
            ):  # This handles lines that don't have expected number format
                print(f"Error parsing line: {line}")
                continue

    return layer_mapping


def read_from_layers_info(filepath: PathType) -> str:
    """Returns a layermap python script from layers.info file."""
    input_text = open(filepath).read()
    layer_mapping = extract_layers(input_text)
    output = "class LayerMap(BaseModel):\n"

    for layer_name, value_tuple in layer_mapping.items():
        output += f"    {layer_name}: Layer = {value_tuple}\n"

    return output



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.typings import Float2, Layers


@gf.cell
def fiducial_squares(
    layers: Layers = ((1, 0),), size: Float2 = (5, 5), offset: float = 0.14
) -> gf.Component:
    """Returns fiducials with two squares.

    Args:
        layers: list of layers.
        size: in um.
        offset: in um.
    """
    c = gf.Component()
    for layer in layers:
        r = c << gf.c.rectangle(size=size, layer=layer)

    for layer in layers:
        r = c << gf.c.rectangle(size=size, layer=layer)
        r.dmove(-np.array(size) - np.array([offset, offset]))

    return c


if __name__ == "__main__":
    c = fiducial_squares()
    c.show()



from __future__ import annotations

import kfactory as kf

from gdsfactory.port import Port


def get_port_x(port: Port) -> float:
    return port.dcenter[0]


def get_port_y(port: Port) -> float:
    return port.dcenter[1]


def sort_ports_x(ports: list[Port]) -> list[Port]:
    ports = list(ports)
    f_key = get_port_x
    ports.sort(key=f_key)
    return ports


def sort_ports_y(ports: list[Port]) -> list[Port]:
    ports = list(ports)
    f_key = get_port_y
    ports.sort(key=f_key)
    return ports


def sort_ports(
    ports1: list[Port] | kf.Ports,
    ports2: list[Port] | kf.Ports,
    enforce_port_ordering: bool,
) -> tuple[list[Port], list[Port]]:
    """Returns two lists of sorted ports.

    Args:
        ports1: the starting ports
        ports2: the ending ports
        enforce_port_ordering: if True, only ports2 will be sorted in accordance with ports1.
            If False, the two lists will be sorted independently.

    """
    ports1 = list(ports1)
    ports2 = list(ports2)

    if len(ports1) != len(ports2):
        raise ValueError(f"ports1={len(ports1)} and ports2={len(ports2)} must be equal")
    if not ports1:
        raise ValueError("ports1 is an empty list")
    if not ports2:
        raise ValueError("ports2 is an empty list")

    if ports1[0].orientation in [0, 180] and ports2[0].orientation in [0, 180]:
        _sort(get_port_y, ports1, enforce_port_ordering, ports2)
    elif ports1[0].orientation in [90, 270] and ports2[0].orientation in [90, 270]:
        _sort(get_port_x, ports1, enforce_port_ordering, ports2)
    else:
        axis = "X" if ports1[0].orientation in [0, 180] else "Y"
        f_key1 = get_port_y if axis in {"X", "x"} else get_port_x
        ports1.sort(key=f_key1)
        if not enforce_port_ordering:
            ports2.sort(key=f_key1)

    if enforce_port_ordering:
        ports2 = [ports2[ports1.index(p1)] for p1 in ports1]

    return ports1, ports2


def _sort(key_func, ports1, enforce_port_ordering, ports2):
    ports1.sort(key=key_func)
    if not enforce_port_ordering:
        ports2.sort(key=key_func)


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component()
    c1 = c << gf.c.straight()
    c2 = c << gf.c.straight()
    sort_ports(c1.ports, c2.ports, enforce_port_ordering=True)
    c.show()



import gdsfactory as gf

if __name__ == "__main__":
    c = gf.Component()
    ref1 = c.add_ref(gf.components.rectangle(size=(10, 10), layer=(1, 0)))
    ref2 = c.add_ref(gf.components.text("Hello", size=10, layer=(2, 0)))
    ref3 = c.add_ref(gf.components.text("world", size=10, layer=(2, 0)))

    ref1.dxmax = ref2.dxmin - 5
    ref3.dxmin = ref2.dxmax + 2
    ref3.drotate(30)
    c.show()
